//////////////////////////////////////////////
// main.cpp
//
//////////////////////////////////////////////

/// Includes ---------------------------------

// nkLog
#include <NilkinsLog/ConsoleLogger.h>

// nkGraphics
#include <NilkinsGraphics/Cameras/Camera.h>
#include <NilkinsGraphics/Cameras/CameraManager.h>

#include <NilkinsGraphics/Log/LogManager.h>

#include <NilkinsGraphics/Graph/Entity.h>
#include <NilkinsGraphics/Graph/SubEntity.h>

#include <NilkinsGraphics/Graph/Node.h>
#include <NilkinsGraphics/Graph/NodeManager.h>

#include <NilkinsGraphics/Meshes/Load/MeshLoader.h>

#include <NilkinsGraphics/Meshes/Mesh.h>
#include <NilkinsGraphics/Meshes/MeshManager.h>

// REGION_CPP nkGraphicsProgramInclude
#include <NilkinsGraphics/Programs/Program.h>
#include <NilkinsGraphics/Programs/ProgramManager.h>
#include <NilkinsGraphics/Programs/ProgramSourcesHolder.h>
// !REGION_CPP nkGraphicsProgramInclude

#include <NilkinsGraphics/RenderContexts/RenderContextManager.h>

#include <NilkinsGraphics/RenderQueues/RenderQueue.h>
#include <NilkinsGraphics/RenderQueues/RenderQueueManager.h>

// REGION_CPP nkGraphicsShaderInclude
#include <NilkinsGraphics/Shaders/Memory/ConstantBuffer.h>
#include <NilkinsGraphics/Shaders/Memory/ShaderInstanceMemorySlot.h>
#include <NilkinsGraphics/Shaders/Memory/ShaderPassMemorySlot.h>

#include <NilkinsGraphics/Shaders/Shader.h>
#include <NilkinsGraphics/Shaders/ShaderManager.h>
// !REGION_CPP nkGraphicsShaderInclude

#include <NilkinsGraphics/Systems/MainSystem.h>

// nkResources
#include <NilkinsResources/ResourceManager.h>

// nkWinUi
#include <NilkinsWinUi/Systems/MainSystem.h>

// Standards
#include <chrono>

#include <memory>

/// Internals : Mesh -------------------------

void prepareMeshFromFile ()
{
	nkGraphics::Mesh* mesh = nkGraphics::MeshManager::getInstance()->createOrRetrieve("Triangle") ;

	nkGraphics::MeshLoader::fillBufferFromFile(mesh, "sphere.obj") ;
	mesh->load() ;
}

void addMeshToRenderQueue ()
{
	nkGraphics::RenderQueue* rq = nkGraphics::RenderQueueManager::getInstance()->get(0) ;

	nkGraphics::Mesh* mesh = nkGraphics::MeshManager::getInstance()->createOrRetrieve("Triangle") ;

	nkGraphics::Entity* ent = rq->addEntity() ;
	nkGraphics::SubEntity* subEnt = ent->addChild() ;

	subEnt->setMesh(mesh) ;
}

void addEntityToGraph ()
{
	nkGraphics::RenderQueue* rq = nkGraphics::RenderQueueManager::getInstance()->get(0) ;
	nkGraphics::Entity* ent = rq->getEntity(0) ;

	nkGraphics::Node* node = nkGraphics::NodeManager::getInstance()->create() ;
	node->setPositionAbsolute(nkMaths::Vector(0.f, 0.f, 10.f)) ;

	ent->setParentNode(node) ;
}

/// Internals : Shader -----------------------

void prepareProgram ()
{
	// REGION_CPP nkGraphicsProgramCreate
	nkGraphics::Program* program = nkGraphics::ProgramManager::getInstance()->createOrRetrieve("program") ;
	// !REGION_CPP nkGraphicsProgramCreate

	// REGION_CPP nkGraphicsProgramSources
	nkGraphics::ProgramSourcesHolder sources ;

	sources.setVertexMemory
	(
		R"eos(
			cbuffer PassBuffer : register(b0)
			{
				matrix view ;
				matrix proj ;
			}

			struct VertexInput
			{
				float4 position : POSITION ;
				matrix world : WORLDMAT ;
			} ;

			struct PixelInput
			{
				float4 position : SV_POSITION ;
			} ;

			PixelInput main (VertexInput input)
			{
				PixelInput result ;

				matrix mvp = mul(input.world, mul(view, proj)) ;
				result.position = mul(input.position, mvp) ;

				return result ;
			}
		)eos"
	) ;

	sources.setPixelMemory
	(
		R"eos(
			struct PixelInput
			{
				float4 position : SV_POSITION ;
			} ;
				
			float4 main (PixelInput input) : SV_TARGET
			{
				return float4(0.8, 0.4, 0.2, 1.0) ;
			}
		)eos"
	) ;

	program->setFromMemory(sources) ;
	program->load() ;
	// !REGION_CPP nkGraphicsProgramSources
}

void prepareShader ()
{
	// REGION_CPP nkGraphicsShaderCreation
	nkGraphics::Shader* shader = nkGraphics::ShaderManager::getInstance()->createOrRetrieve("shader") ;
	// !REGION_CPP nkGraphicsShaderCreation

	nkGraphics::Program* program = nkGraphics::ProgramManager::getInstance()->get("program") ;

	// REGION_CPP nkGraphicsShaderSettings
	shader->setAttachedShaderProgram(program) ;

	nkGraphics::ConstantBuffer* cBuffer = shader->addConstantBuffer(0) ;

	nkGraphics::ShaderPassMemorySlot* slot = cBuffer->addPassMemorySlot() ;
	slot->setAsViewMatrix() ;

	slot = cBuffer->addPassMemorySlot() ;
	slot->setAsProjectionMatrix() ;

	nkGraphics::ShaderInstanceMemorySlot* instanceSlot = shader->addInstanceMemorySlot() ;
	instanceSlot->setAsWorldMatrix() ;

	shader->load() ;
	// !REGION_CPP nkGraphicsShaderSettings
}

void assignShader ()
{
	nkGraphics::RenderQueue* rq = nkGraphics::RenderQueueManager::getInstance()->get(0) ;
	nkGraphics::Entity* ent = rq->getEntity(0) ;
	nkGraphics::Shader* shader = nkGraphics::ShaderManager::getInstance()->get("shader") ;

	// REGION_CPP nkGraphicsAssignShader
	ent->setShader(shader) ;
	// !REGION_CPP nkGraphicsAssignShader
}

/// Function ---------------------------------

int main ()
{
	// Prepare for logging
	std::unique_ptr<nkLog::Logger> logger = std::make_unique<nkLog::ConsoleLogger>() ;
	nkGraphics::LogManager::getInstance()->setReceiver(logger.get()) ;

	// For easiness
	nkResources::ResourceManager::getInstance()->setWorkingPath("Data") ;

	// Initialize and create context with window
	if (!nkGraphics::MainSystem::getInstance()->initialize())
		return -1 ;

	nkGraphics::RenderContext* context = nkGraphics::RenderContextManager::getInstance()->createRenderContextAndWindowFromConfig() ;

	// Prepare the mesh we will show
	prepareMeshFromFile() ;
	addMeshToRenderQueue() ;
	addEntityToGraph() ;

	// Change its shader
	prepareProgram() ;
	prepareShader() ;
	assignShader() ;

	// Get some data to prepare for the rendering loop
	nkGraphics::Camera* camera = nkGraphics::CameraManager::getInstance()->getDefaultCam() ;

	nkGraphics::MainSystem* nkGraphicsSystem = nkGraphics::MainSystem::getInstance() ;
	nkWinUi::MainSystem* nkWinUiSystem = nkWinUi::MainSystem::getInstance() ;

	const float loopTimeMs = 5 * 1000.f ;
	std::chrono::system_clock::time_point firstNow = std::chrono::system_clock::now() ;

	const nkMaths::Vector sphereCenter = nkMaths::Vector(0.f, 0.f, 10.f) ;

	// Our rendering loop is ready to start
	while (nkGraphicsSystem->getHasRunToContinue())
	{
		// Frame the graphics system
		if (!nkGraphicsSystem->frame(context))
			break ;

		// Loop the camera around the sphere
		std::chrono::system_clock::time_point now = std::chrono::system_clock::now() ;
		float currentTimeMs = std::chrono::duration_cast<std::chrono::microseconds>(now - firstNow).count() / 1000.f ;
		
		float currentLoopT = (std::fmod(currentTimeMs, loopTimeMs) / loopTimeMs) * 2.f - 1.f ;
		currentLoopT *= 3.14 ;
		
		nkMaths::Vector loopPos (std::cos(currentLoopT) * 10.f, 0.f, std::sin(currentLoopT) * 10.f) ;
		loopPos += sphereCenter ;

		camera->setPositionAbsolute(loopPos) ;
		camera->lookAt(sphereCenter, nkMaths::Vector(0.f, 1.f, 0.f)) ;

		// Tick the windowing system
		nkWinUiSystem->tick() ;
	}

	// Clean exit
	nkGraphics::MainSystem::getInstance()->kill() ;
}